/******************************************************************/
/* filename: cf_rt.c */
/* version : 1.00 */
/* author  : Jan Feenstra */
/******************************************************************/

#include "cp.h"
#include "cfpe.h"
#include "cf_pe.h"
#include "processConfig.h"
#include "debug.h"

/******************************************************************/
/* external declaration of pipes */
extern int CSAP_dataind_pipe[2];
extern int MSAP_dataind_pipe[2];
extern int cf_st2cf_rt_pipe[2];
extern int cf_rt2cf_st_pipe[2];

/******************************************************************/
/* Declaration of external functions */
extern int InsertConfid();
extern int InsertUsertitle();
extern int writeTitle();
extern int writeData();
extern int new_rtst_pdu();
extern int UpdateConference();
extern int showPartners();
extern int showConference();
extern int show_pdu();
extern int show_datareq_mdu();
extern int show_dataind_mdu();
extern int show_cdu();
extern int show_rtst_pdu();

/******************************************************************/
/* function prototying (forward declaration) */
int cf_rt(struct ConfPartnersStruc *pp);
int RemoveKnownConfPartner(struct partner_Type partner[MAXPARTNERS+1], 
  int srcAddr);
int new_join_cdu(struct join_cdu_Type *cdu, char usertitle[USERTITLELENGTH],
  char confid[CONFIDLENGTH]);
int new_datareq_cdu(struct datareq_cdu_Type *cdu, char length,
  char data[DATALENGTH]);
int new_dataind_cdu( struct dataind_cdu_Type *cdu, char size,
  char data[DATALENGTH], char usertitle[USERTITLELENGTH]);
int new_leave_cdu( struct leave_cdu_Type *cdu);
int AddNewPartner( struct partner_Type partner[MAXPARTNERS+1],
  int srcAddress, char usertitle[USERTITLELENGTH]);
int RefConfPartner( struct partner_Type partner[MAXPARTNERS+1],
  int srcAddress);
int IsConfPartner( struct partner_Type partner[MAXPARTNERS+1], 
  int srcAddress);
int IsConfMember( struct ConfPartnersStruc *pp, int srcAddress);
int cmpConfid( char *id1, char *id2);

/******************************************************************/
int cf_rt(struct ConfPartnersStruc *pp)
{
 struct dataind_mdu_Type mdu;
 fd_set read_mask;

 while (1) 
 {
  FD_ZERO(&read_mask);
  FD_SET(MSAP_dataind_pipe[0], &read_mask);
  FD_SET(cf_st2cf_rt_pipe[0], &read_mask);

  select(32, &read_mask, (fd_set *) 0, (fd_set *) 0, (struct timeval *) 0);

  if (FD_ISSET(cf_st2cf_rt_pipe[0], &read_mask))
  {
   struct join_pdu_Type pdu;

   int read_result;
   read_result = read(cf_st2cf_rt_pipe[0], (char *)&pdu, sizeof(struct pdu_Type));
   if(read_result < 0)  
   {  
	Error(Cf_rt,"cf_rt.c: Reading from cf_st2cf_rt_pipe[0]!");
    return(-1);  
   }  
   else if (read_result == 0) 
   {
    return(0); /* EOF */
   }
   
   switch(pdu.type)
   {
    case JOIN:
	{
	 struct join_pdu_Type *ppdu;
	 ppdu = (struct join_pdu_Type *)&pdu;

     Debug(Cf_rt,"cf_rt.c: Entering the 'rtst' join case!");
     /* the user wants to participate a conference or create one! */
	 /* Update partner set! */
     Debug(Cf_rt,"cf_rt.c: Create conference!");
 	 if (!mutant(276)) 
 	 {
	  UpdateConference(&conference, ppdu->usertitle, ppdu->confid, 1/*active*/);
	 }
	 break;
	}
    case LEAVE:
	{
	 struct leave_pdu_Type *ppdu;
	 ppdu = (struct leave_pdu_Type *)&pdu;

     Debug(Cf_rt,"cf_rt.c: Entering the 'rtst' leave case!");
	 /* the user wants to leave the conference */ 
     Debug(Cf_rt,"cf_rt.c: leave conference!");
 	 if (!mutant(358)) 
 	 {
	  UpdateConference(&conference, ppdu->usertitle, ppdu->confid, 0);
	 }
	 /* clear partner list */
     Debug(Cf_rt,"cf_rt.c: Clear partners!");
 	 if (!mutant(294)) 
 	 {
	  conference.partner[0].msapid = 0;
	 }
	 break;
	}
    default:
	{
	 Warning(Cf_pe, "Unexpected data unit from 'cf_st2cf_rt_pipe[0]'");
	 break;
	}
   }
  }
  
  if (FD_ISSET(MSAP_dataind_pipe[0], &read_mask))
  { 
   int read_result;
   read_result = read(MSAP_dataind_pipe[0], (char *)&mdu, sizeof(struct dataind_mdu_Type)); 
   if(read_result < 0)  
   {  
	Error(Cf_rt,"cf_rt.c: Reading from MSAP_dataind_pipe[0]!");
    return(-1);  
   }  
   else if (read_result == 0) 
   {
    return(0); /* EOF */
   }
 
   switch (mdu.data[0])
   {
    case JOIN:
	{
	 struct join_pdu_Type *ppdu;
	 struct rtst_pdu_Type rpdu;

     Debug(Cf_rt,"cf_rt.c: Entering the CSAP join case!");
	 /* Another participant wants to join a named conference! */
	 if (conference.active==0) {
      Debug(Cf_rt,"cf_rt.c: No conference!");
	  break;
	 } 
     Debug(Cf_rt,"cf_rt.c: Conference is active!");
     
     /* Verify if the participant wants to join the same conference! */
	 ppdu = (struct join_pdu_Type *)&mdu.data;
	 if (cmpConfid(ppdu->confid,conference.confid)==0) {
	  Debug(Cf_rt,"cf_rt.c: Sender joins different conference!");
	  break;
	 }
     Debug(Cf_rt,"cf_rt.c: Conference Identifier is equal!");

	 /* Verify, if the sender is allowed to pariticipate? */
 	 if (!mutant(666)) 
 	 {
      if (!IsConfMember(pp,mdu.srcAddress)) 
 	  { 
 	   Debug(Cf_rt,"cf_rt.c: Joining sender is not a member!");
 	   break;
      }
      Debug(Cf_rt,"cf_rt.c: Joining sender is a member!");
	 }
     
     /* Verify if the new partner is already participating */
	 if (IsConfPartner(conference.partner, mdu.srcAddress)) {
	  Debug(Cf_rt,"cf_rt.c: Joining sender is already a partner!");
	 }
	 else
	 {
     /* Insert new partner! */
	  Debug(Cf_rt,"cf_rt.c: Add new partner!");
 	  if (!mutant(467)) 
 	  {
	   AddNewPartner((struct partner_Type *) &conference.partner, \
	    mdu.srcAddress, ppdu->usertitle);
	  }
     } 
	 Debug(Cf_rt,"cf_rt.c: Joining sender is not a partner!");

	 Debug(Cf_rt,"cf_rt.c: Create a rtst answer unit!");
	 /* Ask SendTask to update new partner and send answer-pdu to sender */
	 /* Create rtst_pdu. */
	 new_rtst_pdu(&rpdu, ANSWER,  mdu.srcAddress, ppdu->usertitle, \
	   ppdu->confid);
	   
	 Debug(Cf_rt,"cf_rt.c: Send the  rtst answer unit!");
 	 if (!mutant(247)) 
 	 {
	  write(cf_rt2cf_st_pipe[1], (char *)&rpdu,sizeof(struct rtst_pdu_Type));
	 }
	 break;
	}
	case ANSWER:
	{
	 struct answer_pdu_Type *pdu;
	 struct rtst_pdu_Type rpdu;

     Debug(Cf_rt,"cf_rt.c: Entering the CSAP answer case!");
     /* Another partner replies to your join request! */
	 if (conference.active==0) {
	  Debug(Cf_rt,"cf_rt.c: No conference");
	  break;
     }
	 Debug(Cf_rt,"cf_rt.c: Conference is active!");

	 pdu = (struct answer_pdu_Type *)&mdu.data;
	 if (!mutant(398)) 
 	 {
	  if (cmpConfid(pdu->confid, conference.confid)==0) {
	   Debug(Cf_rt,"cf_rt.c: Conference identifiers are not equal!");
	   break;
      }
	  Debug(Cf_rt,"cf_rt.c: Conference identifier are the same!");
 	 }
     
 	 if (!mutant(293)) 
 	 {
	  if (IsConfPartner(conference.partner, mdu.srcAddress))
	  {
	   Debug(Cf_rt,"cf_rt.c: Sender is already a partner!");
	   Debug(Cf_rt,"cf_rt.c: Keeping old usertitle!");
	   /* We keep using the old usertitle (nickname) because specs says */
	   /* nothing so we do nothing*/
	   break;
      }
	  Debug(Cf_rt,"cf_rt.c: Sender is not a partner!");
	 }
     
 	 if (!mutant(444)) 
 	 {
	  if (!IsConfMember(pp, mdu.srcAddress)) {
	   Debug(Cf_rt,"cf_rt.c: Answering partner is not a member");
	   break;
      }
	  Debug(Cf_rt,"cf_rt.c: Answering partner is a member");
	 }
	 
	 Debug(Cf_rt,"cf_rt.c: Adding answering sender to partner");
     /* Insert sender as a partner */
 	 if (!mutant(289)) 
 	 {
	  AddNewPartner((struct partner_Type *)&conference.partner, \
	    mdu.srcAddress, pdu->usertitle);
	 }

	 Debug(Cf_rt,"cf_rt.c: Create rtst new unit!");
	 /* Create rtst_pdu. */
	 new_rtst_pdu(&rpdu, NEW, mdu.srcAddress, pdu->usertitle, \
	   pdu->confid);

	 Debug(Cf_rt,"cf_rt.c: Sending rtst new unit!");
	 /* Ask SendTask to update new partner! */
 	 if (!mutant(738)) 
 	 {
	  write(cf_rt2cf_st_pipe[1], (char *)&rpdu,sizeof(struct rtst_pdu_Type));
	 }
	 break;
	}
	case DATA:
	{
	 struct rtst_pdu_Type rtst_pdu;
	 struct join_pdu_Type *pdu;
	 struct dataind_cdu_Type cdu;
	 struct data_pdu_Type *ppdu;
	 int refAddr;

     Debug(Cf_rt,"cf_rt.c: Entering the CSAP data case!");
	 /* Someone had sent you a data message */
	 if (conference.active==0) {
	  Debug(Cf_rt,"cf_rt.c: No conference!");
	  break;
     }
     Debug(Cf_rt,"cf_rt.c: Conference is active!");

	 /* Cannot verify if conference  identifier is correct! */

	 /* Verify is sender is a registered member! */
     if (!IsConfMember(pp, mdu.srcAddress))
	 {
	  Debug(Cf_rt,"cf_rt.c: Sender is not a member!");
	  break;
     }
	 Debug(Cf_rt,"cf_rt.c: Sender is a member!");

	 /* verify, if  the sender participates our conference! */
     if (!IsConfPartner(conference.partner, mdu.srcAddress)) 
	 {
	  /* sender is not participating this conference */
	  /* maby an answer-pdu got lost! re-send a join request to sender! */
 	  Debug(Cf_rt,"cf_rt.c: Sender is not a partner!");
 	  
 	  Debug(Cf_rt,"cf_rt.c: Construct a join-pdu!");
      pdu = (struct join_pdu_Type *)&mdu.data;
	  new_rtst_pdu(&rtst_pdu, JOIN, mdu.srcAddress, pdu->usertitle,pdu->confid);
	  
 	  Debug(Cf_rt,"cf_rt.c: Sending a join-pdu to rtst !");
 	  if (!mutant(836)) 
 	  {
	   write(cf_rt2cf_st_pipe[1],(char *)&rtst_pdu,sizeof(struct rtst_pdu_Type));
	  }
	  break;
     }
     Debug(Cf_rt,"cf_rt.c: Sender is a partner!");

     Debug(Cf_rt,"cf_rt.c: Decompose data message!");
 	 /* sender is indeed participating this conference */
	 /* decompose message and deliver it to user! */
	 refAddr = RefConfPartner(conference.partner, mdu.srcAddress);

     Debug(Cf_rt,"cf_rt.c: Construct dataind cdu!");
     ppdu = (struct data_pdu_Type *)&mdu.data;
	 new_dataind_cdu(&cdu,ppdu->length,ppdu->data, \
	   conference.partner[refAddr].usertitle);
	   
     Debug(Cf_rt,"cf_rt.c: Sending dataind cdu to user!");
 	 if (!mutant(384)) 
 	 {
      write(CSAP_dataind_pipe[1], (char *)&cdu, 
	    sizeof(struct dataind_cdu_Type));
	 }
	 break;
	}
	case LEAVE:
	{
	 struct leave_pdu_Type *pdu;
	 struct rtst_pdu_Type rpdu;

     Debug(Cf_rt,"cf_rt.c: Entering the CSAP leave case!");
	 if (conference.active==0) {
	  Debug(Cf_rt,"cf_rt.c: No conference!");
	  break;
	 }
	 Debug(Cf_rt,"cf_rt.c: Conference is active!");

	 pdu = (struct leave_pdu_Type *)&mdu.data;
	 if (cmpConfid(conference.confid, pdu->confid)==0) {
	  Debug(Cf_rt,"cf_rt.c: Conference identifiers are not equal!");
	  break;
	 }
     Debug(Cf_rt,"cf_rt.c: Conference identifiers are equal!");

     if (!IsConfPartner(conference.partner, mdu.srcAddress)) {
	  Debug(Cf_rt,"cf_rt.c: Leaving sender is not a partner!");
	  break;
     }
	 Debug(Cf_rt,"cf_rt.c: Leaving sender is a partner!");
	 
	 Debug(Cf_rt,"cf_rt.c: Removing sender from partner!");
 	 if (!mutant(856)) 
 	 {
      RemoveKnownConfPartner((struct partner_Type *)&conference.partner, \
	   mdu.srcAddress);
	 }

	 Debug(Cf_rt,"cf_rt.c: Construct rtst pdu!");
	 /* Notify the sendtask to remove sender. */
	 new_rtst_pdu(&rpdu, LEAVE, mdu.srcAddress, pdu->usertitle, pdu->confid);
	 
	 Debug(Cf_rt,"cf_rt.c: Sending rtst pdu to sendtask!");
 	 if (!mutant(777)) 
 	 {
	  write(cf_rt2cf_st_pipe[1],(char *)&rpdu,sizeof(struct rtst_pdu_Type));
	 }
	 break;
	}
	default:
	{
	 Warning(Cf_rt,"Unexpected data unit from 'MSAP_dataind_pipe[0]'");
	 break;
    }
   }
  }
   
 }
}
/******************************************************************/

int RemoveKnownConfPartner(struct partner_Type partner[MAXPARTNERS+1], 
  int srcAddr)
{
 int i,j;
 i = RefConfPartner(partner, srcAddr);
	 
 /* find last entry in partner list! */
 j=i+1;
 while (partner[j].msapid!=0) j++;
 j--;
 if (i!=j) {
  partner[i].msapid = partner[j].msapid;
  InsertUsertitle(partner[j].usertitle, &partner[i].usertitle);
  /* It look that it is possible at some time that user j can receive */
  /* the same message more than once! */
 }
 partner[j].msapid = 0;
 return(0);
}
/******************************************************************/

int new_join_cdu(struct join_cdu_Type *cdu, char usertitle[USERTITLELENGTH],
  char confid[CONFIDLENGTH])
{
 cdu->type = JOIN;
 InsertUsertitle(usertitle, &cdu->usertitle);
 InsertConfid(confid, &cdu->confid);
 return(0);
}
/******************************************************************/

int new_datareq_cdu(struct datareq_cdu_Type *cdu, char length,
  char data[DATALENGTH])
{
 cdu->type = DATAREQ;
 cdu->length = length;
 memcpy((char *)&cdu->data, (char *)data, length);
 return(0);
}
/******************************************************************/

int new_dataind_cdu( struct dataind_cdu_Type *cdu, char size,
  char data[DATALENGTH], char usertitle[USERTITLELENGTH])
{
 cdu->type = DATAIND;
 cdu->length = size;
 memcpy((char *)&cdu->data, (char *)data, size);
 InsertUsertitle(usertitle, &cdu->usertitle);
 return(0);
}
/******************************************************************/

int new_leave_cdu( struct leave_cdu_Type *cdu)
{
 cdu->type = LEAVE;
 return(0);
}
/******************************************************************/

int AddNewPartner( struct partner_Type partner[MAXPARTNERS+1],
  int srcAddress, char usertitle[USERTITLELENGTH])
{
 int i=0;
 while (partner[i].msapid!=0) i++;
 partner[i+1].msapid = 0;
 partner[i].msapid = srcAddress;
 InsertUsertitle(usertitle, &partner[i].usertitle);
 return(0);
}
/******************************************************************/

int RefConfPartner( struct partner_Type partner[MAXPARTNERS+1],
  int srcAddress)
{
 int i=0;
 while (partner[i].msapid!=0) 
 {
  if (partner[i].msapid==srcAddress) break;
  i++;
 }
 return(i);
}
/******************************************************************/

int IsConfPartner( struct partner_Type partner[MAXPARTNERS+1], 
  int srcAddress)
{
 int i;
 i = RefConfPartner(partner, srcAddress);

 if (partner[i].msapid==0)
  return(0);
 else
  return(1);
}
/******************************************************************/

int IsConfMember( struct ConfPartnersStruc *pp, int srcAddress)
{
 int j;
 for (j=0; j<pp->size; j++)
  if (pp->Partner[j].msapid==srcAddress) return(1);
 return(0);
}
/******************************************************************/

int cmpConfid( char *id1, char *id2)
{	
 int i;
 for (i=0; i<CONFIDLENGTH; i++)
  if (id1[i]!=id2[i]) break;
 return(i==10);
} 

