/******************************************************************/
/* filename: cp.c */
/* author  : Jan Feenstra */
/******************************************************************/
#include <stdio.h>
#include <signal.h>
#include "debug.h"

static char version[6] = "v2.01b";

/******************************************************************/
/* CSAP pipe declaration */
int CSAP_datareq_pipe[2];
int CSAP_dataind_pipe[2];

/******************************************************************/
/* Task Communication pipes */
/* int exit_mc_rt_pipe[2]; */

/******************************************************************/
/* external declaration of processes */
extern int user(); /* See user.c */
extern int user2(); /* See user2.c */
extern int cfpe(); /* See cfpe.c */

/******************************************************************/
/* function prototying (forward declaration) */
int main(int argc, char *argv[]);

/******************************************************************/
/* Declaration of local variables */
char config[100];
int debug_level  = Main + User; /* Default is 3 = (Main + User) */
extern int mutant_level; /* see mutant.c */

static int PId=0;				/* process identifier */ 

/******************************************************************/
 static void terminate(int rtnvalue)
 {
  if (PId > 0) 
  {
   kill(PId, SIGTERM);
   Debug(Dbg,"cp.c: terminated %d, with return value %d\n", PId, rtnvalue);
   PId = 0;
  }
  exit(0);
 }
/******************************************************************/

int main(int argc, char *argv[])
{
 int narg=argc;
 int idx=1;
 int setInterface=0; /* default interface is USER_INTERFACE */
 
 Info(Main,"Conference Protocol %s\n", version);
 
Debug(Dbg,"1) Value of narg = %d, value of idx = %d\n",narg, idx);
 
 if ((narg>=3) && (strcmp(argv[idx],"-c")==0))
 {
  strcpy((char *)&config, (const char *)argv[idx+1]);
  Info(Main,"Using configuration file: %s\n",config);
  narg -= 2;
  idx += 2;
 }
 else
 {
  Info(Main,"Using configuration file: config.txt\n");
  config[0] = '\0';
 }
Debug(Dbg,"2) Value of narg = %d, value of idx = %d\n",narg, idx);
 if ((narg>=3) && (strcmp(argv[idx],"-DEBUG")==0))
 {
  sscanf((const char *)argv[idx+1], "%d", &debug_level);
  Info(Main,"Using Debug value: %d\n", debug_level);
  narg -= 2;
  idx += 2;
 }
 else
 {
  Info(Main,"Using Debug value: %d\n", debug_level);
 }
Debug(Dbg,"3) Value of narg = %d, value of idx = %d\n",narg, idx);
 if ((narg>=2) && (strcmp(argv[idx],"-CSAP")==0))
 {
  setInterface=1;
  Info(Main,"Using the CSAP INTERFACE directly.\n");
  narg--;
  idx++;
 } 
 else
 {
  setInterface=0;
  Info(Main,"Using the USER_INTERFACE.\n");
 }
Debug(Dbg,"4) Value of narg = %d, value of idx = %d\n",narg, idx);
 if ((narg>=3) && (strcmp(argv[idx],"-MUTANT")==0))
 {
  sscanf((const char *)argv[idx+1], "%d", &mutant_level);
  Info(Main,"Using Mutant code: %d\n", mutant_level);
  narg -= 2;
  idx += 2;
 }
 else
 {
  Info(Main,"Using Mutant code: %d\n", mutant_level);
 }
Debug(Dbg,"5) Value of narg = %d, value of idx = %d\n",narg, idx);

 
 /* cp_init means creation the CSAP pipes for user() and cfpe() */
 if (pipe(CSAP_datareq_pipe)!=0) 
 {
  Error(Main,"cp.c: Creating CSAP_datareq_pipe\n");
  return(-1);
 }
 if (pipe(CSAP_dataind_pipe)!=0)
 {
  Error(Main,"cp.c: Creating CSAP_dataind_pipe\n");
  /* Note that CSAP_datareq_pipe isn't closed! */
  close(CSAP_datareq_pipe[0]);
  close(CSAP_datareq_pipe[1]);
  return (-1);
 }

 /* Set desired termination signals */
 signal(SIGHUP, terminate);
 signal(SIGTERM, terminate);
 signal(SIGQUIT, terminate);
 signal(SIGIOT, terminate);
 signal(SIGCHLD, terminate);
 
 if ((PId = fork()) == -1)
 {
  Error(Main,"cp.c: Executing fork()\n");
  return(-1);
 }

 if (PId)
 {
  int rvalue;
  /* close the pipes which the user() does not use! */

  close(CSAP_datareq_pipe[0]);
  close(CSAP_dataind_pipe[1]);

  /* Startup the user() */
  if (setInterface==0) 
  {
Debug(Dbg,"Entering the user-interface process\n");
    rvalue = user();
  }
  else
  {
Debug(Dbg,"Entering the CSAP user-interface process\n");
    rvalue = user2();
  }
  terminate(rvalue);
  Debug(Dbg,"cp.c: Exit user() detected...\n");
  
  /* close other pipes */
  close(CSAP_datareq_pipe[1]);
  close(CSAP_dataind_pipe[0]);

  return rvalue;
 }
 else
 {
  int rvalue;

  /* close the pipes which the cfpe process does not use! */
  close(CSAP_datareq_pipe[1]);
  close(CSAP_dataind_pipe[0]);

  /* startup cfpe() */
  rvalue = cfpe(&config);
  terminate(rvalue);
  Debug(Dbg,"cp.c: Exit cfpe() detected...");

  /* close the other pipes */
  close(CSAP_datareq_pipe[0]);
  close(CSAP_dataind_pipe[1]);

  return rvalue;
 }
}

